/* 
**  mod_clamav.c -- Apache module to virus check everything returned to clients
**
**  (c) 2003 Dr. Andreas Mueller, Beratung und Entwicklung
**
**  $Id: mod_clamav.c,v 1.8 2003/11/20 15:04:02 afm Exp $
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_core.h"
#include "ap_config.h"
#include "apr_strings.h"
#include "apr_time.h"
#include <clamav.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include "mod_clamav_version.h"
#include <alloca.h>

#ifndef CLAMAV_DEBUG
#define	CLAMAV_DEBUG	0
#endif /* CLAMAV_DEBUG */
#if CLAMAV_DEBUG
#define	clamav_ap_log_rerror(x, args...) ap_log_rerror(x, ##args)
#else
#define clamav_ap_log_rerror(args...)
#endif

#define	MOD_CLAMAV_LOCAL	0
#define	MOD_CLAMAV_DAEMON	1

module AP_MODULE_DECLARE_DATA clamav_module;

/* the clamav statistics context */
#define	VIRUS_LIST_LENGTH	10
#define VIRUS_NAME_LENGTH	64
#define	VIRUS_URI_LENGTH	256
#define	VIRUS_REQ_LENGTH	64

typedef struct clamav_virusinfo {
    pid_t	pid;
    char	uri[VIRUS_URI_LENGTH];
    char	req[VIRUS_REQ_LENGTH];
    char	virus[VIRUS_NAME_LENGTH];
    off_t	size;
    apr_time_t	when;
} clamav_virusinfo;

typedef struct clamav_stats {
    unsigned long	requests;
    unsigned long	checked;
    unsigned long	aborted;
    unsigned long long	totalsize;
    off_t		maxsize;
    off_t		maxvirussize;
    double		cpu;
    unsigned long	viruses;
    clamav_virusinfo	lastviruses[VIRUS_LIST_LENGTH];
    int			last;
    unsigned long	reloads;
} clamav_stats;

static void	clamav_virus_record(clamav_stats *stats, const char *uri,
		    const char *req, const char *virus, off_t size) {
    /* find the index of the entry we are about to create */
    if (++stats->last >= VIRUS_LIST_LENGTH) stats->last = 0;
    stats->lastviruses[stats->last].pid = getpid();
    stats->lastviruses[stats->last].when = apr_time_now();
    stats->lastviruses[stats->last].size = size;

    /* reset the string fields */
    memset(stats->lastviruses[stats->last].uri, 0, VIRUS_URI_LENGTH);
    memset(stats->lastviruses[stats->last].req, 0, VIRUS_REQ_LENGTH);
    memset(stats->lastviruses[stats->last].virus, 0, VIRUS_NAME_LENGTH);

    /* remember uri and virus name */
    strncpy(stats->lastviruses[stats->last].uri, uri, VIRUS_URI_LENGTH - 1);
    strncpy(stats->lastviruses[stats->last].req, req, VIRUS_REQ_LENGTH - 1);
    strncpy(stats->lastviruses[stats->last].virus, virus, VIRUS_NAME_LENGTH -1);
}

/* we don't want to duplicate the clamav trie, so we keep it in a separate */
/* structure								   */
typedef struct cl_daemon_t {
    struct sockaddr	*s;
    int			len;
    int			domain;
} cl_daemon_t;

typedef struct cl_local_t {
    struct cl_node	*root;
    time_t		lastreload;
} cl_local_t;

typedef union clamav_common {
    cl_local_t	*local;
    cl_daemon_t	*daemon;
} clamav_common;

/* the clamav module configuration					*/
typedef struct clamav_config_rec {
    char		*tmpdir;
    int			mode;		/* mod_clamav mode: local or daemon */
    char		*dbdir;		/* local: where are virus patterns */
    int			port;		/* daemon: clamd port number */
    char		*socket;	/* daemon: unix domain socket path */
    struct cl_limits	lim;		/* clamav limits structure */
    int			trickle_interval;
    int			trickle_size;
    int			sizelimit;
    int			reload_interval;
    apr_table_t		*safetypes;	/* types safe to bypass av checking */
    /* virus scanning engine */
    clamav_common	*common;
    /* shared data for statistics */
    apr_shm_t		*shm;
    clamav_stats	*stats;
    /* mutex to mediate access to shared data */
    apr_proc_mutex_t	*mutex;
} clamav_config_rec;

/* the clamav filter context						*/
typedef struct clamav_ctx {
    apr_bucket_brigade	*bb;
    char		*filename;
    int			file;
    time_t		last_trickle;
    int			trickle_offset;
    clamav_config_rec	*module;
    int			bypass;
    int			bytes;
} clamav_ctx;

/* get the module configuration record */
static clamav_config_rec	*clamav_get_module_config(request_rec *r) {
    clamav_config_rec	*result;

    /* make sure we were not handed an empty request, since we don't	*/
    /* know the server, we cannot log anywhere using Apache's API	*/
    if (NULL == r) {
	fprintf(stderr, "[%d] NULL request handed to clamav_get_module_config",
	    (int)getpid());
	return NULL;
    }

    /* try to recover the directory configuration record from the	*/
    /* the apache runtime						*/
    result = (clamav_config_rec *)ap_get_module_config(
	r->per_dir_config, &clamav_module);
    if (NULL != result) {
	return result;
    }

    /* if we get to this point, then there is no module config		*/
    /* record around. This really should not happen, but it did		*/
    /* an a SuSE Linux 9.0 System with the prefork MT			*/
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "[%d] no config record",
	(int)getpid());
    return NULL;
}

/* functions to lock and unlock */
static void	clamav_lock(clamav_config_rec *rec) {
    if (NULL != rec->mutex)
	apr_proc_mutex_lock(rec->mutex);
}
static void	clamav_unlock(clamav_config_rec *rec) {
    if (NULL != rec->mutex)
	apr_proc_mutex_unlock(rec->mutex);
}

/* function to decide whether it is safe to bypass virus checking */
static int	clamav_safe_to_bypass(ap_filter_t *f) {
    clamav_config_rec	*rec;
    int			rc = 1, i, l;
    char		*ct;

    /* get the module record						*/
    rec = clamav_get_module_config(f->r);
    if (NULL == rec) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] no config, assuming safe", (int)getpid());
	goto safe;
    }

    /* only GET and POST return something to the caller, so all others	*/
    /* can safely bypass the virus checker				*/
    if (strcasecmp(f->r->method, "GET") && strcasecmp(f->r->method, "POST"))
	goto safe;

    /* if we don't know the type, we cannot bypass			*/
    if (NULL == f->r->content_type)
	goto unsafe;

    /* find the type							*/
    ct = (char *)apr_pstrdup(f->r->pool, f->r->content_type);
    l = strlen(ct);
    for (i = 0; i < l; i++) ct[i] = tolower(ct[i]);

    /* if the type is among the safe types, we are safe			*/
    if (apr_table_get(rec->safetypes, ct))
	goto safe;
unsafe:
    rc = 0;
safe:
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] bypass for %s is %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri,
	(rc) ? "safe" : "not safe");
    return rc;
}

/* initialize the engine (depends on mode) */
static int	clamav_init_local(ap_filter_t *f, clamav_config_rec *rec) {
    int	ret;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] local init",
	(int)getpid());
    /*  make sure that rec->common->local is allocated			*/
    if (NULL == rec->common->local) {
	rec->common->local = (cl_local_t *)malloc(sizeof(cl_local_t));
	rec->common->local->root = NULL;
	rec->common->local->lastreload = 0;
    }

    /* make sure we know in which directory to look for databases	*/
    if (NULL == rec->dbdir) {
	rec->dbdir = cl_retdbdir();
    }

    /* reload the database from the directory				*/
    if (0 != (ret = cl_loaddbdir(rec->dbdir, &rec->common->local->root, NULL))){
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot initialize clamav engine: %s", (int)getpid(),
	    cl_perror(ret));
	return -1;
    }
    if (rec->shm) {
	clamav_lock(rec);
	rec->stats->reloads++;
	clamav_unlock(rec);
    }
    time(&rec->common->local->lastreload); 	/* remember when reloaded */
    cl_buildtrie(rec->common->local->root);

    /* if we get to this point, we were successfull initializing the local*/
    /* scanner								*/
    return 0;
}

static int	clamav_init_daemon(ap_filter_t *f, clamav_config_rec *rec) {
/* XXX - alloc should be from server pool */
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] daemon init",
	(int)getpid());
    rec->common->daemon = (cl_daemon_t *)malloc(sizeof(cl_daemon_t));
    if (rec->socket) {
	/* initialize daemon connections via a Unix domain socket	*/
	struct sockaddr_un	*sau;
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] init unix domain socket '%s'", (int)getpid(), rec->socket);
	sau = (struct sockaddr_un *)malloc(sizeof(struct sockaddr_un));
	rec->common->daemon->s = (struct sockaddr *)sau;
	rec->common->daemon->len = sizeof(struct sockaddr_un);
	rec->common->daemon->domain = AF_UNIX;
	/* initialize unix domain socket address */
	/* sau->sun_len = strlen(rec->socket); */
	sau->sun_family = PF_UNIX;
	strncpy(sau->sun_path, rec->socket, sizeof(sau->sun_path));
    } else {
	/* initialize daemon connections via a TCP socket		*/
	struct sockaddr_in	*sai;
	sai = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	rec->common->daemon->s = (struct sockaddr *)sai;
	rec->common->daemon->len = sizeof(struct sockaddr_in);
	rec->common->daemon->domain = AF_INET;
	/* initialize INET domain socket address */
	/* sai->sin_len = sizeof(sai->sin_addr); */
	sai->sin_family = PF_INET;
	sai->sin_port = htons((short)rec->port);
	sai->sin_addr.s_addr = htonl(0x7f000001);
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] init inet socket to %s:%d", (int)getpid(),
	    inet_ntoa(sai->sin_addr), rec->port);
    }
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] daemon init complete", (int)getpid());
    return 0;
}

/* initialize the virus scanning engine: either the library inside the
   binary, or using the daemon (prefered) */
static int	clamav_engine_init(ap_filter_t *f, clamav_config_rec *rec) {
    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	return clamav_init_local(f, rec);
	break;
    case MOD_CLAMAV_DAEMON:
	return clamav_init_daemon(f, rec);
	break;
    default:
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] unknown mode. %d",
	    (int)getpid(), rec->mode);
    }
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] working in mode %d",
	(int)getpid(), rec->mode);
    return  -1;
}

/* check whether the engine is ready. In the case of the local scanner,
   reloads the database if necessary as a side effect */
static int	clamav_engine_ready(ap_filter_t *f, clamav_config_rec *rec) {
    time_t	now;
    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	/* the module is not read if the local structure has never been	*/
	/* allocated							*/
	if (NULL == rec->common->local)
	    return 0;
	/* even if the local structure is allocated, it may be the case	*/
	/* that the reload interval has passed, so we may have to reload*/
	/* the database							*/
	if (0 == rec->reload_interval)
	    return 1;	/* always ready, since we are not going to ever	*/
			/* reload the pattern database			*/
	time(&now);
	if ((now - rec->common->local->lastreload) > rec->reload_interval) {
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
		"[%d] database reload necessary after %d seconds",
		(int)getpid(), (int)(now - rec->common->local->lastreload));
	    /* make sure the engine is properly deallocated		*/
	    cl_freetrie(rec->common->local->root);
	    rec->common->local->root = NULL;
	    return 0;
	}
	return 1;
	break;
    case MOD_CLAMAV_DAEMON:
	return (NULL != rec->common->daemon);
	break;
    }
    /* no other state is possible, but how should the compiler know	*/
    /* that? so if we get a different state, will just tell the caller	*/
    /* that we aren't ready, which isn't that wrong after all		*/
    return 0;
}

/* create HTML page informing about virus */
static int	clamav_virus_info(ap_filter_t *f, char *virname) {
    char	infopage[2048];
    clamav_ctx	*ctx = (clamav_ctx *)f->ctx;
    /* set content type header to text/html */
    f->r->content_type = "text/html";      

    /* format page output */
    if (!f->r->header_only) {
	snprintf(infopage, sizeof(infopage),
		"<html>\n"
		"<head>\n"
		"<title>Virus found</title>\n"
		"</head>\n"
		"<body>\n"
		"<h1>Virus found</h1>\n"
		"<p><tt>mod_clamav</tt> version %s found the virus</p>"
		"<p><pre><b>%s</b></pre></p>"
		"<p>while downloading <i>%s</i>."
		"The transfer has been aborted.</p>\n"
		"<hr />&copy; mod_clamav %s\n"
		"</body>\n"
		"</html>\n",
	    mod_clamav_version, virname,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri, mod_clamav_version);
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] sending %d bytes to client", (int)getpid(), strlen(infopage));
	ap_fwrite(f->next, ctx->bb, infopage, strlen(infopage));
    }
    /* this never happens, but we want the compiler to be happy		*/
    return 0;
}

/* perform a virus scan */
static int	clamav_scanfile(ap_filter_t *f, char **virname,
	apr_size_t *len) {

    int			s, l, bytes;
    char		*repl, *code, *v;
    clamav_ctx		*ctx = (clamav_ctx *)f->ctx;
    clamav_config_rec	*rec = ctx->module;
    FILE		*fp;
    char		scancmd[2048];

    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	/* virus scan using local clamav library */
	return cl_scanfile(ctx->filename, virname, (unsigned long *)len,
	    ctx->module->common->local->root, &ctx->module->lim, CL_ARCHIVE);
	break;
    case MOD_CLAMAV_DAEMON:
	/* set reasonable defaults for virname */
	*virname = "daemon connection problem"; *len = 25;

	/* connect to daemon */
	s = socket(rec->common->daemon->domain, SOCK_STREAM, 0);
	if (s < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot create socket: %s (%d)", (int)getpid(),
		strerror(errno), errno);
	    return CL_VIRUS;
	}
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] daemon socket created: %d", (int)getpid(), s);
	if (connect(s, rec->common->daemon->s, rec->common->daemon->len) < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot connect: %s (%d)", (int)getpid(), strerror(errno),
		errno);
	    close(s);
	    return CL_VIRUS;
	}
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] connected to daemon", (int)getpid());

	/* send scanning request */
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] sending scan command for %s", (int)getpid(), ctx->filename);
	bytes = snprintf(scancmd, sizeof(scancmd), "SCAN %s\n", ctx->filename);
	if (bytes != write(s, scancmd, bytes)) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] failed to write scan command: %s", (int)getpid(),
		strerror(errno));
	}

	/* it seems to be easier to read the reply using stdio, so we
	   fdopen the socket */
	if (NULL == (fp = fdopen(s, "r"))) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot fdopen: %s (%d)", (int)getpid(), strerror(errno),
		errno);
	    close(s);
	    return CL_VIRUS;
	}

	/* read until we meet end of file */
	l = strlen(ctx->filename) + 1024;
	repl = (char *)apr_palloc(f->r->pool, l);
	memset(repl, 0, l);
	if (repl != fgets(repl, l, fp)) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot read reply: %s (%d)", (int)getpid(),
		strerror(errno), errno);
	    fclose(fp);
	    return CL_VIRUS;
	}
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] got scan reply: %s", (int)getpid(), (repl) ? repl : "(null)");
	fclose(fp);

	/* elementary checks on the reply */
	if (repl == NULL) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] NULL scan reply", (int)getpid());
	    return CL_VIRUS;
	}

	/* check that the reply string is not empty. Since we'll be using
	   a pointer to the last character later, we compute it now */
	l = strlen(repl) - 1;
	if (l < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] short scan reply %d", (int)getpid(), l);
	    return CL_VIRUS;
	}

	/* as long as we are inside the string, and have \n and \r characters
	   at the end, we keep removing them */
	while ((l >= 0) && ((repl[l] == '\n') || (repl[l] == '\r'))) {
	    repl[l] = '\0';	/* remove the last character */
	    l--;		/* go to the last but one character */
	}

	/* last word: FOUND or ERROR */
	code = strrchr(repl, ' ');
	if (code == repl) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] reply has "
		"only one word: %s\n", (int)getpid(), repl);
	    return CL_VIRUS;
	} 
	*code = '\0'; code++;
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] reply code: %s",
	    (int)getpid(), code);
	if (0 == strcmp(code, "OK"))
	    return CL_CLEAN;
	if (0 == strcmp(code, "ERROR"))
	    return CL_VIRUS;
	if (0 == strcmp(code, "FOUND")) {
	    /* last but one word: virus name */
	    v = strrchr(repl, ' ');
	    if (v == NULL) {
		*virname = "unknown virus";
	    } else {
		*virname = (char *)apr_pstrdup(f->r->pool, v + 1);
	    }
	    *len = strlen(*virname);
	    return CL_VIRUS;
	}

	/* if nothing bad has happend, then the file must be clean 	*/
	return CL_CLEAN;
	break;
    }
    /* this should not happen, but if it does, accept the file 		*/
    return CL_CLEAN;
}

/* initialize the virus filter */
static int	clamav_initialize(ap_filter_t *f) {
    clamav_config_rec	*rec;
    clamav_ctx		*ctx;
    char		*tmpfilename;

    /* get the module record						*/
    rec = clamav_get_module_config(f->r);
    if (NULL == rec) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] no dir config for clamav", (int)getpid());
	return APR_ENODIR;
    }

    /* initialize the clamav filter context				*/
    if (NULL == f->ctx) {
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] ctx alloc required, rec = %x", (int)getpid(),
	    (unsigned int)rec);

	/* increment the request counter				*/
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] mutex at %x", (int)getpid(), (unsigned int)rec->mutex);
	if (rec->shm) {
	    clamav_lock(rec);
	    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] stats at %x", (int)getpid(), (unsigned int)rec->stats);
	    rec->stats->requests++;
	    clamav_unlock(rec);
	}

	/* allocate a new context structure 				*/
	f->ctx = apr_pcalloc(f->r->pool, sizeof(clamav_ctx));
	ctx = (clamav_ctx *)f->ctx;
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] ctx alloced", (int)getpid());

	/* find out whether it is safe to bypass this request, if so,	*/
	/* we don't need the rest of the structure and can safely skip	*/
	/* its initialization						*/
	if (0 != (ctx->bypass = clamav_safe_to_bypass(f)))
	    return APR_SUCCESS;
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] not bypassed", (int)getpid());

	/* create additional data structures				*/
	ctx->bb = apr_brigade_create(f->r->pool, f->c->bucket_alloc);
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] new output brigade created", (int)getpid());

	ctx->module = rec;
	ctx->last_trickle = time(NULL);
	ctx->trickle_offset = 0;
	ctx->bytes = 0;

	/* allocate a temporary filename				*/
	tmpfilename = tempnam((rec->tmpdir) ? rec->tmpdir : "/tmp", "clam");
	ctx->filename = apr_pstrdup(f->r->pool, tmpfilename);
	free(tmpfilename);
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] temporary filename is %s", (int)getpid(), ctx->filename);

	/* we open the file read/write, so we can read from the file	*/
	/* without a need to reopen it					*/
	ctx->file = open(ctx->filename, O_CREAT|O_TRUNC|O_RDWR, 0666);
	if (ctx->file < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot open temporary file: %s", (int)getpid(),
		strerror(errno));
	}
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] method %s request for %s", (int)getpid(),
	    (NULL == f->r->method) ? "(null)" : f->r->method,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);
    } else {
#if 0
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] context already initialized at %x", (int)getpid(),
	    (unsigned int)f->ctx);
#endif
    }

    /* check for an initialized clamav scan engine, if it is 		*/
    /* initialized, we can skip its initialization			*/
    if (clamav_engine_ready(f, rec))
	return APR_SUCCESS;
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] clamav engine initialization required", (int)getpid());

    /* initialize Clamav engine						*/
    clamav_engine_init(f, rec);
    /* if we get to this point, we can be sure that the clamav engine	*/
    /* has been configured correctly					*/

    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] initialization of clamav engine complete", (int)getpid());
    return APR_SUCCESS;
}

/* The content handler */
static void	clamav_display_virus(request_rec *r, clamav_virusinfo *v) {
    char		when[32];
    apr_size_t		ret;
    apr_time_exp_t	t;
    /* if pid is zero, the record is not initialized */
    if (v->pid == 0) return;

    ap_rputs("  <tr>\n", r);
    ap_rprintf(r, "    <td>%d</td>\n", (int)(v->pid));
    apr_explode_localtime(&t, v->when);
    apr_strftime(when, &ret, sizeof(when), "%d %b %Y %H:%M:%S", &t);
    ap_rprintf(r, "    <td>%s</td>\n", when);
    ap_rprintf(r, "    <td>%s</td>\n", v->req);
    ap_rprintf(r, "    <td>%s</td>\n", v->virus);
    ap_rprintf(r, "    <td>%s</td>\n", v->uri);
    ap_rprintf(r, "    <td align=\"right\">%.0f</td>\n", (double)(v->size));
    ap_rputs("  </tr>\n", r);
}

/*
 * clamav_handler
 *
 * the clamav statistics content handler
 */
static int clamav_handler(request_rec *r) {
    clamav_config_rec	*rec;
    int			i, from;

    /* get the module record						*/
    rec = clamav_get_module_config(r);

    /* initialize the clamav filter context				*/
    if (strcmp(r->handler, "clamav")) {
        return DECLINED;
    }
    r->content_type = "text/html";      

    if (!r->header_only) {
	/* lock the shared memory */
	clamav_lock(rec);
	ap_rputs("<html>", r);
	ap_rputs("<head><title>Clamav Module Statistics</title></head>\n", r);
	ap_rputs("<body bgcolor=\"#ffffff\">\n", r);
	ap_rputs("<h1>Clamav Module Status</h1>\n", r);
	ap_rputs("<h2>Configuration</h2>\n", r);
	ap_rputs("<table>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Version:</td><td>%s</td>\n", mod_clamav_version);
	ap_rputs("  </tr>\n", r);
	ap_rputs("  <tr>\n", r);
	if (rec->mode == MOD_CLAMAV_LOCAL) {
	    ap_rputs("    <td>Scanner:</td><td>local</td>\n", r);
	} else {
	    if (rec->socket)
		ap_rprintf(r, "    <td>Daemon Unix Socket:</td><td>%s</td>\n",
		    rec->socket);
	    else
		ap_rprintf(r, "    <td>Daemon TCP Socket:</td>"
		    "<td>127.0.0.1:%d</td>\n", rec->port);
	}
	ap_rputs("  </tr>\n", r);
	ap_rputs("</table>\n", r);
	ap_rputs("<h2>Statistics</h2>\n", r);
	if (rec->shm) {
	    ap_rputs("<table>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Total requests:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->requests);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Checked for viruses:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->checked);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Total data checked:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>kB</td>\n",
		(double)(rec->stats->totalsize)/1024.);
	    ap_rputs("  </tr>\n", r);
	    if (rec->stats->checked > 0) {
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Average size of checked requests:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%.f</td><td>bytes</td>\n",
		    (double)rec->stats->totalsize/(double)rec->stats->checked);
		ap_rputs("  </tr>\n", r);
	    }
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Maximum request size:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>bytes</td>\n",
		(double)(rec->stats->maxsize));
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Maximum virus size:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>bytes</td>\n",
		(double)(rec->stats->maxvirussize));
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Viruses found:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->viruses);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Aborted:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->aborted);
	    ap_rputs("  </tr>\n", r);
	    if (rec->mode == MOD_CLAMAV_LOCAL) {
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Total CPU time:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%.3f</td><td>sec</td>\n",
		    rec->stats->cpu);
		ap_rputs("  </tr>\n", r);
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Database reloads:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		    rec->stats->reloads);
		ap_rputs("  </tr>\n", r);
	    }
	    ap_rputs("</table>\n", r);

	    ap_rprintf(r, "<h2>Last %d viruses found:</h2>\n",
		VIRUS_LIST_LENGTH);

	    ap_rputs("<table border=\"1\">\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>PID</td>\n", r);
	    ap_rputs("    <td>When</td>\n", r);
	    ap_rputs("    <td>Requestor</td>\n", r);
	    ap_rputs("    <td>Virus</td>\n", r);
	    ap_rputs("    <td>Request URI</td>\n", r);
	    ap_rputs("    <td align=\"right\">Size</td>\n", r);
	    ap_rputs("  </tr>\n", r);

	    /* tail portion of the list */
	    from = rec->stats->last + 1;
	    if (from < 1) from = 1;
	    for (i = from; i < VIRUS_LIST_LENGTH; i++)
		clamav_display_virus(r, &rec->stats->lastviruses[i]);

	    /* front portion of the list */
	    for (i = 0; i <= rec->stats->last; i++)
		clamav_display_virus(r, &rec->stats->lastviruses[i]);

	    ap_rputs("</table>\n", r);
	} else {
	    ap_rputs("<p>No statistics available</p>\n", r);
	}

	ap_rputs("<hr />\n", r);
	ap_rputs("&copy; 2003 Dr. Andreas M&uuml;ller, "
	    "Beratung und Entwicklung\n", r);
	ap_rputs("</body>\n", r);
	ap_rputs("</html>\n", r);
	clamav_unlock(rec);
    }
    return OK;
}

/*
 * clamav_cleanup
 *
 * remove temperary files etc
 */
static void	clamav_cleanup(ap_filter_t *f) {
    clamav_ctx	*ctx = (clamav_ctx *)f->ctx;

    /* when done, remove the temporary file 				*/
    if (NULL != ctx->filename) {
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] unlinking %s", (int)getpid(), ctx->filename);
	close(ctx->file);
	unlink(ctx->filename); ctx->filename = NULL;
    }
}

/*
 * clamav_send_trickle
 *
 * send a byte every now and then, so that the client does not turn
 * impatient
 */
static apr_status_t	clamav_sendtrickle(ap_filter_t *f) {
    time_t		now;
    clamav_ctx		*ctx;
    int			file, bytes;
    char		*str;
    apr_status_t	rc;

    ctx = (clamav_ctx *)f->ctx;
    time(&now);
    if ((now - ctx->last_trickle) < ctx->module->trickle_interval)
	return APR_SUCCESS;
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] trickle time for request %s has arrived", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);

    /* even if we fail to send the trickle, we just remember the	*/
    /* failure								*/
    ctx->last_trickle = now;

    /* allocate a buffer on the stack which is certailnly large enough	*/
    str = (char *)alloca(ctx->module->trickle_size);

    /* read a suitable number of bytes from the file			*/
    if ((file = open(ctx->filename, O_RDONLY)) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot open file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	return APR_SUCCESS;
    }
    if (lseek(file, ctx->trickle_offset, SEEK_SET) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot seek file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	close(file);
	return APR_SUCCESS;
    }
    if (0 >= (bytes = read(file, str, ctx->module->trickle_size))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot read file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	close(file);
	return APR_SUCCESS;
    }

    /* send it to the client as one bucket followed by a flush bucket	*/
    rc = ap_fwrite(f->next, ctx->bb, str, bytes);
    ctx->trickle_offset += bytes;
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending %d trickle bytes @ %d", (int)getpid(), bytes,
	ctx->trickle_offset);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_flush_create(f->c->bucket_alloc));
    if (rc != APR_SUCCESS) {
	ap_pass_brigade(f->next, ctx->bb);
    } else {
	rc = ap_pass_brigade(f->next, ctx->bb);
    }
    if (f->r->connection->aborted)
	rc = APR_ECONNABORTED;
    if (rc != APR_SUCCESS)
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "[%d] sending trickle failed: %d", (int)getpid(), rc);
    return rc;
}

/*
 * clamav_flush
 *
 * auxiliary function to close the conection to the client
 */
static apr_status_t	clamav_flush(ap_filter_t *f) {
    clamav_ctx		*ctx = (clamav_ctx *)f->ctx;
    apr_status_t	rc;

    /* send a flush bucket tot he client 				*/
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending eos bucket for request %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_flush_create(f->c->bucket_alloc));
    return ap_pass_brigade(f->next, ctx->bb);
}

/*
 * clamav_endfile
 *
 * auxiliary function to close the conection to the client
 */
static apr_status_t	clamav_endfile(ap_filter_t *f) {
    clamav_ctx		*ctx = (clamav_ctx *)f->ctx;
    apr_status_t	rc;

    /* send a flush bucket tot he client 				*/
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending eos bucket for request %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_eos_create(f->c->bucket_alloc));
    rc = ap_pass_brigade(f->next, ctx->bb);

    /* after this we will certainly no longer need the temporary file	*/
    clamav_cleanup(f);

    return rc;
}

/*
 * clamav_cpu
 *
 * auxiliary function to compute the cpu time consumed for virus scanning
 */
static double	clamav_cpu(struct rusage *begin, struct rusage *end) {
    long	sec, usec;
    double	cpu;
    sec = end->ru_utime.tv_sec - begin->ru_utime.tv_sec;
    usec = end->ru_utime.tv_usec - begin->ru_utime.tv_usec;
    cpu = sec + (0.000001 * usec);
    sec = end->ru_stime.tv_sec - begin->ru_stime.tv_sec;
    usec = end->ru_stime.tv_usec - begin->ru_stime.tv_usec;
    return cpu + sec + (0.0000001 * usec);
}

/*
 * clamav_record_aborted
 *
 * auxiliary function to handle an abortet connection
 */
static void	clamav_record_aborted(ap_filter_t *f) {
    clamav_ctx	*ctx = (clamav_ctx *)f->ctx;

    ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, f->r,
	"[%d] client requesting %s has disconnected", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    if (ctx->module->shm) {
	clamav_lock(ctx->module);
	ctx->module->stats->aborted++;
	clamav_unlock(ctx->module);
    }
    clamav_cleanup(f);
}

/*
 * clamav_check_file
 *
 * check the file we already have for viruses
 */
static apr_status_t	clamav_check_file(ap_filter_t *f) {
    char		*virname;
    struct rusage	begin, end;
    int			len, ret;
    clamav_ctx		*ctx = (clamav_ctx *)f->ctx;
    struct stat		sb;
    off_t		size = 0;
    
    /* perform virus scan						*/
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] initiating virus check in file %s", (int)getpid(),
	ctx->filename);
    virname = NULL;
    getrusage(RUSAGE_SELF, &begin);
    ret = clamav_scanfile(f, &virname, &len);
    getrusage(RUSAGE_SELF, &end);

    /* log the virus checking achievement				*/
    if (stat(ctx->filename, &sb) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot stat file %s: %s (%d)", (int)getpid(),
	    ctx->filename, strerror(errno), errno);
    } else {
	size = sb.st_size;
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] virus checked %s (%s), %.0f bytes, in %.0fms",
	    (int)getpid(), (NULL == f->r->uri) ? "(null)" : f->r->uri,
	    f->r->content_type, (double)size,
	    1000. * clamav_cpu(&begin, &end));
    }

    /* perform the statistics update					*/
    if (ctx->module->shm) {
	clamav_lock(ctx->module);
	ctx->module->stats->cpu += clamav_cpu(&begin, &end);
	ctx->module->stats->checked++;
	ctx->module->stats->totalsize += size;
	if (ctx->module->stats->maxsize < size)
	    ctx->module->stats->maxsize = size;
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] statistics: %d+%d bytes, %d checked", (int)getpid(),
	    (int)ctx->module->stats->totalsize, (int)size,
	    (int)ctx->module->stats->checked);
	clamav_unlock(ctx->module);
     }

    /* if we have found a virus, we also have to register it in the 	*/
    /* statistics 							*/
    if (CL_VIRUS == ret) {
	int	is_ip;
	/* remember this virus */
	if (ctx->module->shm) {
	    clamav_lock(ctx->module);
	    ctx->module->stats->viruses++;
	    if (ctx->module->stats->maxvirussize < size)
		ctx->module->stats->maxvirussize = size;
	    clamav_virus_record(ctx->module->stats,
		(NULL == f->r->uri) ? "(null)" : f->r->uri,
		ap_get_remote_host(f->r->connection,
		    f->r->per_dir_config, REMOTE_NAME, &is_ip),
		    virname, size);
	    clamav_unlock(ctx->module);
	}

	/* log the virus we have found */
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] virus %s found in request %s", (int)getpid(),
	    virname,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);

	/* if we have not sent any data to the client yet, we can
	   generate a HTML page informing about the virus */
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] trickle bytes sent so far: %d", (int)getpid(),
		ctx->trickle_offset);
	if (ctx->trickle_offset == 0) {
	    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"sending virus info to client");
	    clamav_virus_info(f, virname);
	}
    }

    /* return the result of the virus scanner, so we know what has	*/
    /* happend from the caller						*/
    return ret;
}

/*
 * clamav_send_file
 *
 * send all the data we have in the temporary file to the client, and
 * clean the temporary file up
 */
static apr_size_t	clamav_send_file(ap_filter_t *f) {
    clamav_ctx		*ctx;
    int			bytes;
    apr_size_t		len = 0;
    size_t		totalsize = 0;
    apr_status_t	rc = APR_SUCCESS;
    char		buffer[2048];
    struct stat		sb;

    ctx = (clamav_ctx *)f->ctx;
    /* send the remaining data as bunch of buckets to the, this is done	*/
    /* as follows							*/
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending file contents to client", (int)getpid());

    /* find out how many bytes there currently are in the file		*/
    if (fstat(ctx->file, &sb) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot stat file %s", (int)getpid(), ctx->filename);
    } else {
	totalsize = sb.st_size;
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] expect %d bytes",
	    (int)getpid(), totalsize);
    }

    /* seek to the current trickle position 				*/
    if (lseek(ctx->file, ctx->trickle_offset, SEEK_SET) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot seek to %d: %s (%d)\n", (int)getpid(),
	    ctx->trickle_offset, strerror(errno), errno);
    }

    /* while we can write to the client, and while there is data, read	*/
    /* a buffer full of data an write it to the client			*/
    while ((rc == APR_SUCCESS) &&
	(bytes = read(ctx->file, buffer, sizeof(buffer))) > 0) {
	/* log if there is a difference between what we asked for and	*/
	/* what we got							*/
	if (bytes != sizeof(buffer))
		clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] read %d bytes from %s", (int)getpid(), bytes,
	    ctx->filename);
	len += bytes;
	rc = ap_fwrite(f->next, ctx->bb, buffer, bytes);
    }
    if (bytes < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] read from descriptor %d failed: %s (%d)", (int)getpid(),
	    ctx->file, strerror(errno), errno);
    }
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] file %s written to client (%d bytes)", (int)getpid(),
	ctx->filename, len);

    /* send a flush down the filter chaing				*/
    clamav_flush(f);

    /* check whether we have seen all the file				*/
    if (totalsize > 0) {
	if ((totalsize - ctx->trickle_offset) != len) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, f->r,
		"[%d] not all the file sent to the client: %d != %d\n",
		(int)getpid(), totalsize - ctx->trickle_offset, len);
	} else {
	    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] sent size matches totalsize %d", (int)getpid(), len);
	}
    }

    /* set the trickle_offset to the current offset, as if we had sent	*/
    /* all the stuff as trickles					*/
    ctx->trickle_offset = totalsize;

    /* clean up the temporary file					*/
    clamav_cleanup(f);

    /* return the number of bytes sent to the client 			*/
    return len;
}

/*
 * clamav_save_block
 *
 * utility function to write a block to the temporary file
 */
static int	clamav_save_block(ap_filter_t *f, const char *str, int len) {
    clamav_ctx	*ctx;
    int		bytes;

    ctx = (clamav_ctx *)f->ctx;
    /* make sure to position the file at the end */
    if (lseek(ctx->file, 0, SEEK_END) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] cannot seek to eof",
	    (int)getpid());
	return 0;
    }

    /* now write the block to the file					*/
    bytes = write(ctx->file, str, len);
    if (bytes != len) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] error during write to %s", (int)getpid(),
	    ctx->filename);
    }
    ctx->bytes += bytes;
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] %d bytes written to tmp file %s", (int)getpid(), bytes,
	ctx->filename);
    return bytes;
}

/*
 * clamav_filter
 *
 * the apache2 filter function for clamav based virus scanning
 */
static apr_status_t	clamav_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
    clamav_ctx		*ctx;
    const char		*str;
    int			bytes;
    apr_size_t		len = 81920;
    apr_bucket		*e;
    apr_status_t	rc = APR_SUCCESS;

#if 0
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] clamav_filter for uri %s, ctx = %x", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri,
	(unsigned int)f->ctx);
#endif
    
    /* initialize (if necessary) the virus filter			*/
    if (APR_SUCCESS != (rc = clamav_initialize(f))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "clamav initialization failed");
	return rc;
    }
    ctx = (clamav_ctx *)f->ctx;

    /* if it is safe to bypass the virus checker, just pass the bucket	*/
    /* brigade to the next filter					*/
    if (ctx->bypass) {
	return ap_pass_brigade(f->next, bb);
    }

    /* every data bucket that comes in must be written to the file	*/
    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] scanning bucket brigade", (int)getpid());
    APR_BRIGADE_FOREACH(e, bb) {
	clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] bucket type: %s (uri %s)", (int)getpid(), e->type->name,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);
	/* read the data in the bucket					*/
	apr_bucket_read(e, &str, &len, APR_NONBLOCK_READ);
	if (len > 0) {
	    /* there really is something to write			*/
	    if (len != (bytes = clamav_save_block(f, str, len))) {
		    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
			"[%d] %d bytes written to tmp file %s",
			(int)getpid(), bytes, ctx->filename);
	    }
	}

	/* should give EOS and FLUSH buckets special treatment. But as	*/
	/* noted below, Flush buckets don't really help, so we just drop*/
	/* them, not without a debug log message though			*/
	if ((!APR_BUCKET_IS_EOS(e)) && (!APR_BUCKET_IS_FLUSH(e))) {
	    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] data bucket found", (int)getpid());
	    apr_bucket_delete(e); /* removes any non-EOS or non-FLUSH buckets */
	} else {
	    clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] bucket of type %s found", (int)getpid(),
		APR_BUCKET_IS_EOS(e) ? "End of Stream" : "Flush");
	    if (APR_BUCKET_IS_FLUSH(e)) {
		apr_bucket_delete(e); /* removes  any FLUSH buckets */
	    }
	}
	/* at this point, all non-EOS buckets have been removed		*/
    }

    /* now check whether we should send a trickle byte, but sending	*/
    /* a trickle may fail caused by the client having aborted the 	*/
    /* the connection, so we return an error				*/
    if (APR_SUCCESS != clamav_sendtrickle(f)) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "[%d] trickle detects connection abort", (int)getpid());
	clamav_record_aborted(f);
	return APR_ECONNABORTED;
    }

    /* if we have read up to the size limit already, then we check	*/
    /* the piece we have and send it to the client, and switch to	*/
    /* bypass mode							*/
    if ((ctx->module->sizelimit > 0) && (ctx->module->sizelimit <= ctx->bytes)){
	/* perform the check of the part we already have		*/
	if (CL_VIRUS == clamav_check_file(f)) {
	    clamav_endfile(f);
	    return APR_EGENERAL;
	}

	/* send the file to the client					*/
	clamav_send_file(f);

	/* switch to bypass mode					*/
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "bypassing check of %s due to size limit %d",
	    f->r->uri, ctx->module->sizelimit);
	ctx->bypass = 1;

	/* but we should hand over the bucket brigade to the next	*/
	/* as if we had been bypassing already at the beginning		*/
	return ap_pass_brigade(f->next, bb);
    }

    /* at this point we only have EOS buckets left			*/
    APR_BRIGADE_FOREACH(e, bb) {
	/* if we have met an EOS bucket, we should trigger the virus	*/
	/* scanning							*/
	if (APR_BUCKET_IS_EOS(e)) {
	    /* check for viruses and update the statistics		*/
	    if (CL_VIRUS == clamav_check_file(f)) {
		/* close the connection to the client */
		clamav_endfile(f);
		return APR_EGENERAL;
	    }

	    /* send the remaining data as bunch of buckets to the 	*/
	    /* client							*/
	    clamav_send_file(f);
	    rc = clamav_endfile(f);
	    if (rc != APR_SUCCESS)
		clamav_record_aborted(f);
	    return rc;
	} else {
	    /* for the record: we ignore FLUSH buckets, they don't help	*/
	    /* to detect that the connection was aborted		*/
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] non-EOS bucket found when only EOS buckets should be "
		"left", (int)getpid());
	}
	/* for the record: we ignore FLUSH buckets, they don't help to	*/
	/* detect that the connection was aborted			*/
    }
    apr_brigade_cleanup(bb);
    return APR_SUCCESS;
}

static void clamav_register_hooks(apr_pool_t *p)
{
    ap_hook_handler(clamav_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_register_output_filter("CLAMAV", clamav_filter, NULL,
	AP_FTYPE_CONTENT_SET);
}

/* configuration methods */
static void	*clamav_create_dir_config(apr_pool_t *p, char *d) {
    apr_status_t	rc;
    clamav_config_rec	*cfg;
    cfg = (clamav_config_rec *)apr_pcalloc(p, sizeof(clamav_config_rec));
    cfg->mode = MOD_CLAMAV_LOCAL;
    cfg->dbdir = NULL;
    cfg->tmpdir = NULL;
    cfg->port = 0;
    cfg->common = apr_pcalloc(p, sizeof(clamav_common));
    cfg->common->local = NULL;
    cfg->trickle_interval = 60;
    cfg->trickle_size = 1;
    cfg->reload_interval = 0;
    cfg->sizelimit = 0;
    cfg->lim.maxfiles = 100;
    cfg->lim.maxfilesize = 10 * 1048576;
    cfg->lim.maxreclevel = 8;
    cfg->safetypes = apr_table_make(p, 0);
    cfg->mutex = NULL;
    cfg->shm = NULL;

    /* create the mutex for shared memory */
    if (APR_SUCCESS != apr_proc_mutex_create(&cfg->mutex, "clamav.lock",
	APR_LOCK_DEFAULT, p)) {
	fprintf(stderr, "%s:%d: cannot create mutex clamav.lock",
	    __FILE__, __LINE__);
	cfg->mutex = NULL;
    }

    /* create share memory */
    clamav_lock(cfg);
    rc = apr_shm_attach(&cfg->shm, "clamav.shm", p);
    if (APR_SUCCESS != rc) {
	/* simply attaching to the shared memory didn't work so we must	*/
	/* create it first						*/
	rc = apr_shm_create(&cfg->shm, sizeof(clamav_stats), "clamav.shm", p);
	if (APR_SUCCESS != rc) {
	    fprintf(stderr, "%s:%d: cannot create shared memory clamav.shm: "
		"statistics will not be available\n", __FILE__, __LINE__);
	    cfg->shm = NULL;
	} else {
	    /* since we are the creator of the shared memory segment,	*/
	    /* we must also initialize it				*/
	    cfg->stats = (clamav_stats *)apr_shm_baseaddr_get(cfg->shm);
	    memset(cfg->stats, 0, sizeof(clamav_stats));
	    cfg->stats->cpu = 0.;
	}
    } else {
	/* initialize the pointer to the shm structure			*/
	cfg->stats = (clamav_stats *)apr_shm_baseaddr_get(cfg->shm);
    }
    clamav_unlock(cfg);

    return cfg;
}

static void	*clamav_merge_dir_config(apr_pool_t *p, void *pp, void *cp) {
    clamav_config_rec	*parent = (clamav_config_rec *)pp;
    clamav_config_rec	*child = (clamav_config_rec *)cp;

    /* allways share the common part					*/
    /* XXX what is the correct way to free memory that was apr_palloced?
    if (child->common != parent->common) {
	apr_free(child->common);
    }
    */
    child->common = parent->common;
    child->mode = parent->mode;
    if (parent->dbdir)
	child->dbdir = (char *)apr_pstrdup(p, parent->dbdir);
    if (parent->socket)
	child->socket = (char *)apr_pstrdup(p, parent->socket);
    if (parent->reload_interval > 0)
	child->reload_interval = parent->reload_interval;
    child->trickle_size = parent->trickle_size;
    child->trickle_interval = parent->trickle_interval;
    child->sizelimit = parent->sizelimit;

    /* merge rest (except shared memory and mutex, which are not 	*/
    /* shared)								*/
    if (child->port != parent->port) child->port = parent->port;
    if (child->lim.maxfiles == 0)
	child->lim.maxfiles = parent->lim.maxfiles;
    if (child->lim.maxfilesize == 0)
	child->lim.maxfilesize = parent->lim.maxfilesize;
    if (child->lim.maxreclevel == 0)
	child->lim.maxreclevel = parent->lim.maxreclevel;
    if ((!child->tmpdir) && (parent->tmpdir))
	child->tmpdir = (char *)apr_pstrdup(p, parent->tmpdir);
    if (apr_is_empty_table(child->safetypes))
	apr_table_overlay(p, parent->safetypes, child->safetypes);

    /* return child configuration					*/
    return child;
}

/* add safe types to the list of safe types */
static const char	*clamav_add_safetype(cmd_parms *parms, void *mconfig,
    const char *arg) {
    char	*argi;
    int		i, l;

    /* get the module configuration record				*/
    clamav_config_rec	*rec = (clamav_config_rec *)mconfig;

    /* convert the type to lower case					*/
    argi = apr_pstrdup(parms->server->process->pconf, arg);
    l = strlen(argi);
    for (i = 0; i < l; i++) argi[i] = tolower(argi[i]);
    apr_table_set(rec->safetypes, argi, "safe");
    return NULL;
}

/* mode configuration parsing function */
static const char	*clamav_set_mode_cmd(cmd_parms *parms, void *mconfig,
	const char *arg) {
    clamav_config_rec	*cfg = (clamav_config_rec *)mconfig;

    if (0 == strcasecmp("daemon", arg)) {
	cfg->mode = MOD_CLAMAV_DAEMON;
	return NULL;
    }
    if (0 == strcasecmp("local", arg)) {
	cfg->mode = MOD_CLAMAV_LOCAL;
	return NULL;
    }
    return "unknown Clamav Mode value";
}

/* configuration directives */
static const command_rec	clamav_cmds[] = {
    AP_INIT_TAKE1(
	"ClamavMode", clamav_set_mode_cmd,
	(void *)APR_OFFSETOF(clamav_config_rec, mode),
	RSRC_CONF,
	"virus scanning mode: local or daemon"
    ),
    AP_INIT_TAKE1(
	"ClamavSocket", ap_set_string_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, socket),
	RSRC_CONF,
	"Socket file name if clamd uses Unix domain socket"
    ),
    AP_INIT_TAKE1(
	"ClamavDbdir", ap_set_string_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, dbdir),
	RSRC_CONF,
	"Directory containing clamav databases"
    ),
    AP_INIT_TAKE1(
	"ClamavTmpdir", ap_set_string_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, tmpdir),
	RSRC_CONF,
	"Directory for temporary files during virus checking"
    ),
    AP_INIT_TAKE1(
	"ClamavReloadInterval", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, reload_interval),
	RSRC_CONF,
	"Database reload interval"
    ),
    AP_INIT_TAKE1(
	"ClamavTrickleInterval", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, trickle_interval),
	RSRC_CONF,
	"Trickle interval"
    ),
    AP_INIT_TAKE1(
	"ClamavTrickleSize", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, trickle_size),
	RSRC_CONF,
	"Trickle size"
    ),
    AP_INIT_TAKE1(
	"ClamavMaxfiles", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, lim.maxfiles),
	RSRC_CONF,
	"Maximum number of files in an archive"
    ),
    AP_INIT_TAKE1(
	"ClamavMaxfilesize", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, lim.maxfilesize),
	RSRC_CONF,
	"Maximum archive size"
    ),
    AP_INIT_TAKE1(
	"ClamavRecursion", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, lim.maxreclevel),
	RSRC_CONF,
	"Maximum recursion depth when analyzing archives"
    ),
    AP_INIT_TAKE1(
	"ClamavPort", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, port),
	RSRC_CONF,
	"Port number where clamd can be reached"
    ),
    AP_INIT_ITERATE(
	"ClamavSafetypes", clamav_add_safetype,
	(void *)APR_OFFSETOF(clamav_config_rec, safetypes),
	RSRC_CONF,
	"a list of MIME types for which it is safe to bypass the virus checker"
    ),
    AP_INIT_ITERATE(
	"ClamavSizelimit", ap_set_int_slot,
	(void *)APR_OFFSETOF(clamav_config_rec, sizelimit),
	RSRC_CONF,
	"files larger than the sizelimit will only be checked up to that point"
    ),
    { NULL }
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA clamav_module = {
    STANDARD20_MODULE_STUFF, 
    clamav_create_dir_config,
                           /* create per-dir    config structures */
    clamav_merge_dir_config,
                           /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    clamav_cmds,           /* table of config file commands       */
    clamav_register_hooks  /* register hooks                      */
};

